/*--------------------------------------------------------------------------*/
/*       Copyright(C) 2003-2008 KONICA MINOLTA SENSING,INC.                 */
/*--------------------------------------------------------------------------*/
// VIMatrix.h: interface of the CVIMatrix class
//
//////////////////////////////////////////////////////////////////////

#if !defined(__VIMATRIX_H__)
#define __VIMATRIX_H__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "VIVector.h"

#ifdef DATAPROCESS_EXPORTS
#define DATAPROCESS_API __declspec(dllexport)
#else
#define DATAPROCESS_API __declspec(dllimport)
#endif



class   DATAPROCESS_API CVIMatrixMN
{
    double* m_mt;
    int m_nM, m_nN;
    
public:
    CVIMatrixMN();
    CVIMatrixMN(int m, int n);
    CVIMatrixMN(int n);
    CVIMatrixMN(const CVIMatrixMN &SourceMN);
    ~CVIMatrixMN();

    BOOL Construct(int m, int n);

    inline CVIMatrixMN& operator = ( const CVIMatrixMN& mt0); // *this = Matrix
    inline CVIMatrixMN  operator + ( const CVIMatrixMN& mt0); // *this + Matrix
    inline CVIMatrixMN  operator - ( const CVIMatrixMN& mt0); // *this * Matrix
	inline CVIMatrixMN	operator * ( const CVIMatrixMN& mt0); // *this * Matrix
	inline CVIVectorND	operator * ( const CVIVectorND& vec); // *this * Vector
    inline CVIMatrixMN  operator * ( const double c);		// *this * constant
    inline CVIMatrixMN  operator / ( const double c);		// *this / constant
    inline int      operator ==( const CVIMatrixMN& v);	// *this == Matrix
    inline int      operator !=( const CVIMatrixMN& v);	// *this != Matrix
    inline double&  operator()(int i, int j);
    inline double   operator()(int i, int j) const;

    inline void     t(CVIMatrixMN& matrix); 
    inline int      RowOrder();
    inline int      ColumOrder();
    inline void     CopyRow(CVIVectorND& u, int i);
    inline void     CopyColum(CVIVectorND& u, int j);
    inline void     Product(CVIMatrixMN& m0, CVIMatrixMN& m1);
	inline void 	GetProductVector(CVIVectorND& vec, CVIVectorND& solution);
    inline double   Inverse(CVIMatrixMN& B);
    inline BOOL     Jacobi(CVIMatrixMN& E, CVIVectorND& EV);  
    inline void 	MPInverse(CVIMatrixMN& B);
  
    inline double&	Element(int index);
    inline double&	Value(int i, int j);
    
protected:
    inline void CVIMatrixMN::Copy(const CVIMatrixMN& SourceMN);

};

///////////////////////////////////////////////////////////////////////////////
// --- |operator| ---

inline CVIMatrixMN&   CVIMatrixMN::operator = ( const CVIMatrixMN& mt0)
{
    for (int i=0; i< m_nM; i++)
    {
        for (int j=0; j < m_nN; j++)
        {
            Value(i,j) = mt0(i,j);
        }
    }
    return *this;
}
inline CVIMatrixMN    CVIMatrixMN::operator + (const CVIMatrixMN& mt0)
{
    CVIMatrixMN tmp(m_nM,m_nN);
    if ( RowOrder()   != mt0.m_nM ) return tmp;
    if ( ColumOrder() != mt0.m_nN ) return tmp;
    for (int i=0; i<m_nM; i++)
    {
        for (int j=0; j<m_nN; j++)
        {
            tmp(i,j) = Value(i,j) + mt0(i,j);
        }
    }
    return tmp;
}
inline CVIMatrixMN    CVIMatrixMN::operator - (const CVIMatrixMN& mt0)
{
    CVIMatrixMN tmp(m_nM,m_nN);
    if ( RowOrder()   != mt0.m_nM ) return tmp;
    if ( ColumOrder() != mt0.m_nN ) return tmp;
    for (int i=0; i<m_nM; i++)
    {
        for (int j=0; j< m_nN; j++)
        {
            tmp(i,j) = Value(i,j) - mt0(i,j);
        }
    }
    return tmp;
}
inline CVIMatrixMN    CVIMatrixMN::operator * (const CVIMatrixMN& mt0)
{
    CVIMatrixMN a(m_nM,m_nM);
    if ( mt0.m_nM != ColumOrder() ) return a;
    if ( mt0.m_nN != RowOrder() ) return a;
    for (int i=0; i<m_nM; i++)
    {
        for (int j=0; j<m_nM; j++)
        {
            for (int k=0; k<mt0.m_nN; k++)
            {
                a(i,j) = a(i,j) + Value(i,k)*mt0(k,j);
            }
        }
    }
    return a;
}
inline CVIVectorND    CVIMatrixMN::operator * ( const CVIVectorND& vec)
{
	int nOrder = vec.nOrder;
    CVIVectorND v( nOrder );
    if ( m_nN != nOrder ) return v;
    for (int i=0; i<RowOrder(); i++)
    {
        for (int j=0; j<ColumOrder(); j++)
        {
            v(i) += Value(i,j)*vec(j);
        }
    }
    return v;
}
inline CVIMatrixMN    CVIMatrixMN::operator * ( const double c)
{
    CVIMatrixMN a(m_nM,m_nN);
    for (int i=0; i<m_nM * m_nN; i++)
    {
        a.Element(i) = Element(i)*c;
    }
    return a;
}
inline CVIMatrixMN    CVIMatrixMN::operator / ( const double c)
{
    CVIMatrixMN a(m_nM,m_nN);
    for (int i=0; i<m_nM * m_nN; i++)
    {
        a.Element(i) = Element(i)/c;
    }
    return a;
}
inline int         CVIMatrixMN::operator ==( const CVIMatrixMN& mt0)
{
    int n = 1;
    for (int i=0; i<m_nM * m_nN; i++)
    {
        n = n * ( Element(i) == mt0.m_mt[i] );
    }
    return n;
}
inline int CVIMatrixMN::operator !=( const CVIMatrixMN& mt0)
{
    int n = 1;
    for (int i=0; i<m_nM * m_nN; i++)
    {
        n = n * ( Element(i) != mt0.m_mt[i] );
    }
    return n;
}
inline double& CVIMatrixMN::operator() (int i, int j)
{
    assert( i>=0 && i< m_nM && j>=0 && j< m_nN);

    return m_mt[ i * m_nN + j ];
}
inline double CVIMatrixMN::operator() (int i, int j) const
{
    assert( i>=0 && i< m_nM && j>=0 && j< m_nN);
    return m_mt[ i * m_nN + j ];
}

// --- |Function| ---

inline void CVIMatrixMN::t(CVIMatrixMN& matrix)
{
    assert( matrix.RowOrder() == ColumOrder() && matrix.ColumOrder() == RowOrder() );

    for (int i=0; i < m_nN; i++)
    {
        for (int j=0; j < m_nM; j++)
        {
            matrix(i, j) = Value(j, i);
        }
    }
}
inline int CVIMatrixMN::RowOrder()
{
    return m_nM;
}
inline int CVIMatrixMN::ColumOrder()
{
    return m_nN;
}
inline void CVIMatrixMN::CopyRow(CVIVectorND& u, int i)
{
    assert( i>=0 && i< m_nN);
    assert( u.Order() == m_nN );

    for (int j=0; j<m_nN; j++)
    {
        u(j) = m_mt[i*m_nN + j];
    }
}
inline void CVIMatrixMN::CopyColum(CVIVectorND& u, int j)
{
    assert( j>=0 && j< m_nM);
    assert( u.Order() == m_nM );

    for (int i=0; i<m_nM; i++)
    {
        u(i) = m_mt[i*m_nM + j];
    }
}
inline void CVIMatrixMN::Product(CVIMatrixMN& m0, CVIMatrixMN& m1)
{
	assert( m0.ColumOrder() == m1.RowOrder() );
    assert( m_nM == m0.RowOrder() );
    assert( m_nN == m1.ColumOrder() );

    for (int i=0; i< m_nM; i++)
    {
        for (int j=0; j < m_nN; j++)
        {
            for (int k=0; k<m0.ColumOrder(); k++)
            {
                Value(i,j) = Value(i,j) + m0(i,k)*m1(k,j);
            }
        }
    }
}
inline void CVIMatrixMN::GetProductVector(CVIVectorND& vec, CVIVectorND& solution)
{
	assert( m_nN == vec.Order() );
	assert( m_nM == solution.Order() );

    for (int i=0; i<RowOrder(); i++)
    {
        for (int j=0; j<ColumOrder(); j++)
        {
            solution(i) += Value(i,j)*vec(j);
        }
    }

}


inline double&    CVIMatrixMN::Value(int i, int j)
{
    assert( i>=0 && i< m_nM && j>=0 && j< m_nN);
    return m_mt[ i * m_nN + j ];
}
inline double&    CVIMatrixMN::Element(int index)
{
    return m_mt[ index ];
}

inline void CVIMatrixMN::Copy(const CVIMatrixMN& SourceMN)
{
    int m,n;
    m       = SourceMN.m_nM ;
    n       = SourceMN.m_nN ;
    m_mt    = NULL          ;
    Construct(m,n);

    for (int i=0; i<m*n; i++)
    {
        m_mt[i] = SourceMN.m_mt[i];
    }
}

///////////////////////////////////////////////////////////////////////////////
class DATAPROCESS_API CVIMatrix2
{
private:
    CVIVector2 m_r[2];

protected:
    inline void Copy(const CVIMatrix2& m);
    inline CVIVector2 Col(int i) const;

public:
    //-----------------------------------------------------------------------
    static CVIMatrix2 Identity;
    static CVIMatrix2 Zero;
    static CVIMatrix2 Unit;

public:
    //-----------------------------------------------------------------------
    inline CVIMatrix2()                                      {                           }
    inline CVIMatrix2(const CVIVector2& r0,const CVIVector2& r1) { m_r[0] = r0; m_r[1] = r1; }
    inline CVIMatrix2(const CVIMatrix2& m)                     { Copy(m);                  }

    //-----------------------------------------------------------------------
    inline double& operator()(int i, int j)       { assert(i>=0 && i<2 && j>=0 && j<2);
                                                               return m_r[i][j]; }
    inline double  operator()(int i, int j) const { assert(i>=0 && i<2 && j>=0 && j<2);
                                                               return m_r[i][j]; }
    
    //-----------------------------------------------------------------------
    inline CVIMatrix2& operator=(const CVIMatrix2& m);
    inline CVIMatrix2& operator+=(const CVIMatrix2& m);
    inline CVIMatrix2& operator-=(const CVIMatrix2& m);
    inline CVIMatrix2& operator*=(double s);
    inline CVIMatrix2& operator/=(double s);
    
    //-----------------------------------------------------------------------
    inline CVIMatrix2 operator+(const CVIMatrix2& m) const;
    inline CVIMatrix2 operator-(const CVIMatrix2& m) const;
    inline CVIMatrix2 operator-() const;
    inline CVIMatrix2 operator*(double s) const;
    inline CVIMatrix2 operator/(double s) const;
    inline CVIMatrix2 operator*(const CVIMatrix2& m) const;
    inline CVIVector2 operator*(const CVIVector2& v) const;
    
    //-----------------------------------------------------------------------
    inline double Det();
    inline CVIMatrix2 Transpose();
    inline double Inverse(CVIMatrix2&);
    
    //-----------------------------------------------------------------------
};
/////////////////////////////////////////////////////////////////////////////

/*-------------------------------------------------------------------------*/
inline void CVIMatrix2::Copy(const CVIMatrix2& m)
{
    m_r[0] = m.m_r[0];  m_r[1] = m.m_r[1];
}

/*-------------------------------------------------------------------------*/
inline CVIVector2 CVIMatrix2::Col(int i) const
{ 
    return CVIVector2(m_r[0][i], m_r[1][i]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2& CVIMatrix2::operator=(const CVIMatrix2& m)
{
    Copy(m);
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2& CVIMatrix2::operator+=(const CVIMatrix2& m)
{
    m_r[0] += m.m_r[0]; m_r[1] += m.m_r[1];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2& CVIMatrix2::operator-=(const CVIMatrix2& m)
{
    m_r[0] -= m.m_r[0]; m_r[1] -= m.m_r[1];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2& CVIMatrix2::operator*=(double s)
{
    m_r[0] *= s;        m_r[1] *= s;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2& CVIMatrix2::operator/=(double s)
{
    double t = 1/s;
    m_r[0] *= t;        m_r[1] *= t;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2 CVIMatrix2::operator+(const CVIMatrix2& m) const
{
    return CVIMatrix2(m_r[0]+m.m_r[0], m_r[1]+m.m_r[1]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2 CVIMatrix2::operator-(const CVIMatrix2& m) const
{
    return CVIMatrix2(m_r[0]-m.m_r[0], m_r[1]-m.m_r[1]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2 CVIMatrix2::operator-() const
{
    return CVIMatrix2(-m_r[0], -m_r[1]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2 CVIMatrix2::operator*(const CVIMatrix2& m) const
{
    CVIMatrix2 A;
    for(int i=0;i<2;i++)
	    for(int j=0;j<2;j++)
	        A(i,j) = m_r[i]*m.Col(j);
    return A;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2 CVIMatrix2::operator*(double s) const
{
    return CVIMatrix2(m_r[0]*s, m_r[1]*s);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2 CVIMatrix2::operator/(double s) const
{
    double t = 1/s;
    return CVIMatrix2(m_r[0]*t, m_r[1]*t);
}

/*-------------------------------------------------------------------------*/
inline CVIVector2 CVIMatrix2::operator*(const CVIVector2& v) const
{
    return CVIVector2(m_r[0]*v, m_r[1]*v);
}

/*-------------------------------------------------------------------------*/
inline double CVIMatrix2::Det()
{
    return (m_r[0][0]*m_r[1][1] - m_r[0][1]*m_r[1][0]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix2 CVIMatrix2::Transpose()
{
    return CVIMatrix2(Col(0), Col(1));
}

/*-------------------------------------------------------------------------*/
inline double CVIMatrix2::Inverse(CVIMatrix2& inv)
{
    double d = Det();
    if( d==0.0 )    return 0.0;
    double t = 1/d;
    inv.m_r[0][0] =  m_r[1][1]*t;
    inv.m_r[0][1] = -m_r[1][0]*t;
    inv.m_r[1][0] = -m_r[0][1]*t;
    inv.m_r[1][1] =  m_r[0][0]*t;
    return d;
}


///////////////////////////////////////////////////////////////////////////////

class DATAPROCESS_API CVIMatrix3
{
private:
    CVIVector3 m_r[3];

protected:
    inline void Copy(const CVIMatrix3& m);
    inline CVIVector3 Col(int i) const;
    
public:
    //-----------------------------------------------------------------------
    static CVIMatrix3 Identity;
    static CVIMatrix3 Zero;
    static CVIMatrix3 Unit;
    static CVIMatrix3 Diag(const CVIVector3& v);
    static CVIMatrix3 Extend(const CVIVector3& v);

public:
    //-----------------------------------------------------------------------
    inline CVIMatrix3()                  {                                       }
    inline CVIMatrix3(const CVIVector3& r0,const CVIVector3& r1,const CVIVector3& r2)
                                       { m_r[0] = r0; m_r[1] = r1; m_r[2] = r2;}
    inline CVIMatrix3(const CVIMatrix3& m) { Copy(m);                              }

    //-----------------------------------------------------------------------
    inline double& operator()(int i, int j)       { assert(i>=0 && i<3 && j>=0 && j<3);
                                                               return m_r[i][j]; }
    inline double  operator()(int i, int j) const { assert(i>=0 && i<3 && j>=0 && j<3);
                                                               return m_r[i][j]; }
    
    //-----------------------------------------------------------------------
    inline CVIMatrix3& operator=(const CVIMatrix3& m);
    inline CVIMatrix3& operator+=(const CVIMatrix3& m);
    inline CVIMatrix3& operator-=(const CVIMatrix3& m);
    inline CVIMatrix3& operator*=(double s);
    inline CVIMatrix3& operator/=(double s);
    
    //-----------------------------------------------------------------------
    inline CVIMatrix3 operator+(const CVIMatrix3& m) const;
    inline CVIMatrix3 operator-(const CVIMatrix3& m) const;
    inline CVIMatrix3 operator-() const;
    inline CVIMatrix3 operator*(double s) const;
    inline CVIMatrix3 operator/(double s) const;
    inline CVIMatrix3 operator*(const CVIMatrix3& m) const;
    inline CVIVector3 operator*(const CVIVector3& v) const;
    
    //-----------------------------------------------------------------------
    inline double Det();
    inline CVIMatrix3 Transpose();
    inline CVIMatrix3 Adjoint();
    inline double Inverse(CVIMatrix3&);
    
    inline BOOL Jacobi(CVIMatrix3&, CVIVector3&);
    
};
/////////////////////////////////////////////////////////////////////////////

/*-------------------------------------------------------------------------*/
inline void CVIMatrix3::Copy(const CVIMatrix3& m)
{
    m_r[0] = m.m_r[0];  m_r[1] = m.m_r[1];  m_r[2] = m.m_r[2];
}

/*-------------------------------------------------------------------------*/
inline CVIVector3 CVIMatrix3::Col(int i) const
{
    return CVIVector3(m_r[0][i], m_r[1][i], m_r[2][i]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3& CVIMatrix3::operator=(const CVIMatrix3& m)
{
    Copy(m);
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3& CVIMatrix3::operator+=(const CVIMatrix3& m)
{
    m_r[0] += m.m_r[0]; m_r[1] += m.m_r[1]; m_r[2] += m.m_r[2];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3& CVIMatrix3::operator-=(const CVIMatrix3& m)
{
    m_r[0] -= m.m_r[0]; m_r[1] -= m.m_r[1]; m_r[2] -= m.m_r[2];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3& CVIMatrix3::operator*=(double s)
{
    m_r[0] *= s;        m_r[1] *= s;        m_r[2] *= s;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3& CVIMatrix3::operator/=(double s)
{
    double t = 1/s;
    m_r[0] *= t;        m_r[1] *= t;        m_r[2] *= t;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3 CVIMatrix3::operator+(const CVIMatrix3& m) const
{
    return CVIMatrix3(m_r[0]+m.m_r[0], m_r[1]+m.m_r[1], m_r[2]+m.m_r[2]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3 CVIMatrix3::operator-(const CVIMatrix3& m) const
{
    return CVIMatrix3(m_r[0]-m.m_r[0], m_r[1]-m.m_r[1], m_r[2]-m.m_r[2]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3 CVIMatrix3::operator-() const
{
    return CVIMatrix3(-m_r[0], -m_r[1], -m_r[2]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3 CVIMatrix3::operator*(const CVIMatrix3& m) const
{
    CVIMatrix3 A;
    for(int i=0;i<3;i++)
	    for(int j=0;j<3;j++)
	        A(i,j) = m_r[i]*m.Col(j);
    return A;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3 CVIMatrix3::operator*(double s) const
{
    return CVIMatrix3(m_r[0]*s, m_r[1]*s, m_r[2]*s);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3 CVIMatrix3::operator/(double s) const
{
    double t = 1/s;
    return CVIMatrix3(m_r[0]*t, m_r[1]*t, m_r[2]*t);
}

/*-------------------------------------------------------------------------*/
inline CVIVector3 CVIMatrix3::operator*(const CVIVector3& v) const
{
    return CVIVector3(m_r[0]*v, m_r[1]*v, m_r[2]*v);
}

/*-------------------------------------------------------------------------*/
inline double CVIMatrix3::Det()
{
    return m_r[0]*(m_r[1]^m_r[2]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3 CVIMatrix3::Transpose()
{
    return CVIMatrix3(Col(0), Col(1), Col(2));
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix3 CVIMatrix3::Adjoint()
{
    return CVIMatrix3(m_r[1]^m_r[2], m_r[2]^m_r[0], m_r[0]^m_r[1]);
}

/*-------------------------------------------------------------------------*/
inline double CVIMatrix3::Inverse(CVIMatrix3& inv)
{
    CVIMatrix3 A = Adjoint();
    double d = A.m_r[0] * m_r[0];

    if( d==0.0 )    return 0.0;
    inv = A.Transpose()/d;
    return d;
}


/*-------------------------------------------------------------------------*/

class DATAPROCESS_API CVIMatrix4
{
private:
    CVIVector4 m_r[4];

protected:
    inline void Copy(const CVIMatrix4& m);
    inline CVIVector4 Col(int i) const;
    
public:
    //-----------------------------------------------------------------------
    static CVIMatrix4 Identity;
    static CVIMatrix4 Zero;
    static CVIMatrix4 Unit;
    
    static CVIMatrix4 Trans(double x, double y, double z);
    static CVIMatrix4 Scale(double x, double y, double z);
    static CVIMatrix4 Xrot(double a);
    static CVIMatrix4 Yrot(double a);
    static CVIMatrix4 Zrot(double a);
    
public:
    //-----------------------------------------------------------------------
    inline CVIMatrix4()                             {                           }
    inline CVIMatrix4(const CVIVector4& r0,const CVIVector4& r1,const CVIVector4& r2,const CVIVector4 r3)
                        { m_r[0] = r0; m_r[1] = r1; m_r[2] = r2; m_r[3] = r3; }
    inline CVIMatrix4(const CVIMatrix4& m)            { Copy(m);                  }

    //-----------------------------------------------------------------------
    inline double& operator()(int i, int j)       { assert(i>=0 && i<4 && j>=0 && j<4);
                                                               return m_r[i][j]; }
    inline double  operator()(int i, int j) const { assert(i>=0 && i<4 && j>=0 && j<4);
                                                               return m_r[i][j]; }
    inline const CVIVector4& operator[](int i) const { assert(i>=0 && i<4); return m_r[i]; }
    
    //-----------------------------------------------------------------------
    inline BOOL operator==(const CVIMatrix4&);
    
    //-----------------------------------------------------------------------
    inline CVIMatrix4& operator=(const CVIMatrix4& m);
    inline CVIMatrix4& operator+=(const CVIMatrix4& m);
    inline CVIMatrix4& operator-=(const CVIMatrix4& m);
    inline CVIMatrix4& operator*=(double s);
    inline CVIMatrix4& operator/=(double s);
    
    //-----------------------------------------------------------------------
    inline CVIMatrix4 operator+(const CVIMatrix4& m) const;
    inline CVIMatrix4 operator-(const CVIMatrix4& m) const;
    inline CVIMatrix4 operator-() const;
    
    inline CVIMatrix4 operator*(double s) const;
    inline CVIMatrix4 operator/(double s) const;
    inline CVIMatrix4 operator*(const CVIMatrix4& m) const;
    inline CVIVector4 operator*(const CVIVector4& v) const;
    inline CVIVector3 operator*(const CVIVector3& v) const;
    
    //-----------------------------------------------------------------------
    inline double Det() const;
    inline CVIMatrix4 Transpose() const;
    inline CVIMatrix4 Adjoint() const;
    inline double Inverse(CVIMatrix4&) const;
    inline double CramerInverse(CVIMatrix4&) const;

    inline BOOL Jacobi(CVIMatrix4&, CVIVector4&) const;
    
};
/////////////////////////////////////////////////////////////////////////////

/*-------------------------------------------------------------------------*/
inline void CVIMatrix4::Copy(const CVIMatrix4& m)
{
    m_r[0] = m.m_r[0];  m_r[1] = m.m_r[1];  m_r[2] = m.m_r[2];  m_r[3] = m.m_r[3];
}

/*-------------------------------------------------------------------------*/
inline CVIVector4 CVIMatrix4::Col(int i) const
{
    return CVIVector4(m_r[0][i], m_r[1][i], m_r[2][i], m_r[3][i]);
}

/*-------------------------------------------------------------------------*/
inline BOOL CVIMatrix4::operator==(const CVIMatrix4& m)
{
    return (m_r[0]==m.m_r[0] && m_r[1]==m.m_r[1] && m_r[2]==m.m_r[2] && m_r[3]==m.m_r[3]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4& CVIMatrix4::operator=(const CVIMatrix4& m)
{
    Copy(m);
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4& CVIMatrix4::operator+=(const CVIMatrix4& m)
{
    m_r[0] += m.m_r[0]; m_r[1] += m.m_r[1]; m_r[2] += m.m_r[2]; m_r[3] += m.m_r[3];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4& CVIMatrix4::operator-=(const CVIMatrix4& m)
{
    m_r[0] -= m.m_r[0]; m_r[1] -= m.m_r[1]; m_r[2] -= m.m_r[2]; m_r[3] -= m.m_r[3];
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4& CVIMatrix4::operator*=(double s)
{
    m_r[0] *= s;        m_r[1] *= s;        m_r[2] *= s;        m_r[3] *= s;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4& CVIMatrix4::operator/=(double s)
{
    double t = 1/s;
    m_r[0] *= t;        m_r[1] *= t;        m_r[2] *= t;        m_r[3] *= t;
    return *this;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4 CVIMatrix4::operator+(const CVIMatrix4& m) const
{
    return CVIMatrix4(m_r[0]+m.m_r[0], m_r[1]+m.m_r[1], m_r[2]+m.m_r[2], m_r[3]+m.m_r[3]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4 CVIMatrix4::operator-(const CVIMatrix4& m) const
{
    return CVIMatrix4(m_r[0]-m.m_r[0], m_r[1]-m.m_r[1], m_r[2]-m.m_r[2], m_r[3]-m.m_r[3]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4 CVIMatrix4::operator-() const
{
    return CVIMatrix4(-m_r[0], -m_r[1], -m_r[2], -m_r[3]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4 CVIMatrix4::operator*(const CVIMatrix4& m) const
{
    CVIMatrix4 A;
    for(int i=0;i<4;i++)
	    for(int j=0;j<4;j++)
	        A(i,j) = m_r[i]*m.Col(j);
    return A;
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4 CVIMatrix4::operator*(double s) const
{
    return CVIMatrix4(m_r[0]*s, m_r[1]*s, m_r[2]*s, m_r[3]*s);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4 CVIMatrix4::operator/(double s) const
{
    double t = 1/s;
    return CVIMatrix4(m_r[0]*t, m_r[1]*t, m_r[2]*t, m_r[3]*t);
}

/*-------------------------------------------------------------------------*/
inline CVIVector4 CVIMatrix4::operator*(const CVIVector4& v) const
{
    return CVIVector4(m_r[0]*v, m_r[1]*v, m_r[2]*v, m_r[3]*v);
}

/*-------------------------------------------------------------------------*/
inline CVIVector3 CVIMatrix4::operator*(const CVIVector3& v) const
{
    CVIVector4 u = CVIVector4(v,1);
    double w = m_r[3]*u;

    if(w==0.0)  return CVIVector3(m_r[0]*u,   m_r[1]*u,   m_r[2]*u  );
    else        return CVIVector3(m_r[0]*u/w, m_r[1]*u/w, m_r[2]*u/w);
}

/*-------------------------------------------------------------------------*/
inline double CVIMatrix4::Det() const
{
    return m_r[0]*Cross(m_r[1], m_r[2], m_r[3]);
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4 CVIMatrix4::Transpose() const
{
    return CVIMatrix4(Col(0), Col(1), Col(2), Col(3));
}

/*-------------------------------------------------------------------------*/
inline CVIMatrix4 CVIMatrix4::Adjoint() const
{
    CVIMatrix4 A;

    A.m_r[0] = Cross( m_r[1], m_r[2], m_r[3]);
    A.m_r[1] = Cross(-m_r[0], m_r[2], m_r[3]);
    A.m_r[2] = Cross( m_r[0], m_r[1], m_r[3]);
    A.m_r[3] = Cross(-m_r[0], m_r[1], m_r[2]);
    return A;
}

/*-------------------------------------------------------------------------*/
inline double CVIMatrix4::CramerInverse(CVIMatrix4& B) const
{
    CVIMatrix4 A = Adjoint();
    double d = A.m_r[0] * m_r[0];

    if( d==0.0 )    return 0.0;
    B = A.Transpose()/d;
    return d;
}

///////////////////////////////////////////////////////////////////////////////

#endif
